//       SoundTrack game engine demo by Sam Turner
//	 NCCA Innovations Project 2006
// 	 All code Copyright

//	Requires the FMOD sound library (www.fmod.org), and glut.	



#include <GL/glut.h> 				// uses glut
#include <stdlib.h>


#include "musicFunc.c"    			//music handling functions


#define SIN(x) (   sin( ((x) /180) *3.142)  )   //convert radians to degrees
#define COS(x) (   cos( ((x) /180) *3.142)  )


//program constants
					

#define FASTER 1
#define SLOWER 0

#define FLAT 0
#define ROUND 1
#define ROUNDING 2
#define FLATTENING 3
#define FLATTENING_WAITING 4

#define TRUE 1
#define FALSE 0



// speed player moves up and down regarding the wave physics

#define UPSPEED 0.0003
#define DOWNSPEED 0.001


//distance between each line
#define LINE_DIST 0.5


//offset increment for each frame
#define OFSET_INC 0.5


//number of samples in the freq analysis
#define LEVELS 128


//width of track
#define TRACK_WIDTH 0.7



void upDateShape(void);
void updatePlayer();
void changeToRoundTimer();
void swapTrack();
void rotateZTrack();
void setUpSinCosArray();
void updatePlayerHistory(float XPos, float YPos);
void calcNewPlayerUp();
void addNewLine();
void updateCorner();
void setupLines();
void drawCurrentLine(int currentLine, float xS, float yS, float zS, float xE, float yE, float zE);
void drawLines();
void updateFrame();
void changeSpeed(int speed);
void updatePlayer();
void drawPlayer();
void drawMenu();
void display(void);
void keyUp(int key, int x, int y);
void key(int key, int x, int y);
static void idle(void);



//virutal camera structure
typedef struct cam_t
{
        float lookX;
        float lookY;
        float lookZ;
        
        float eyeX;
        float eyeY;
        float eyeZ;
        
        float upX;
        float upY;
        float upZ;
} cam;



//To cameras initialised
cam oldCam = {0}, newCam = {0};


        
/*structure holding information about each line*/
typedef struct lineItem_t
{
        float red;
        float green;
        float blue;
        float angle;            /*angle relative from previous line*/
        float levels[LEVELS];   /*frequency graph*/
        float roundness;	/*the shape of the line, flat = 0, round = 1*/
        } lineItem;
        

/*array holding every line (in a ring buffer)*/
lineItem lineBuffer[20];

/*start and end pointer to element on the array holing the furthest and closest line*/
int lineSP = 0;
int lineEP = 19;

//gamespeed - how often the game state is refreshed in ms
float gameSpeed = 15;

// variables to globally store the shape of the track
float roundness = 0.0;
float roundHere = 0;

/*ofset of angle and distance to display the track (0 - 1)*/  
float OfsetT = 0;

//music frequency, 44100 is normal rate, lower numbers slow down the music
int musicFreq = 44100;

/*current angle of new lines being created*/
float currentAngle = 0;
float currentAngleAim = 0;


//Current Player information and history
float playerAngle = 0;
float sinAngle = 0.0;
float playerXPos = 0.0;
float previousXPos[10] = {0};
float previousYPos[10] = {0};
int previousLast = 0;
float playerSpeed = 1.0;
float playerUpThrust = 0.0;
float playerYPos = 0.0;

//wave height of line at current player position
float waveHeight = 0.0;


/*keyboard handling*/
int pressingLeft=0;
int pressingRight=0;

//Flag to indicate whether the player is in conact with the surface
int onTrack = 1;

//whether or not to draw the onderside of the track (i.e. when it is 'tube' shaped
int drawingBottom = 1;


//flag indicating the current shape of the track
int trackShape = FLAT;

//control over the rotation of the track when player swaps to the underside
float trackRotationAngle = 0;
int trackRotating = 0;


//Difficulty setting
float difficulty = 0.0;

//pre calaculated sin and cos tables
float sinTable[LEVELS*2];
float cosTable[LEVELS*2];



//finally flip the track upside completely
void swapTrack()
{
           int i;
           
           trackRotationAngle = 0;  //reset the rotation angle
           trackRotating = 0;       //stop the track rotating

           
           channelPlaying= (channelPlaying+1)%2;   	// switch music channel
           if (playerXPos<-1) playerXPos+= 2;		//swap the players side
           else if (playerXPos>1) playerXPos-= 2;   
           
           for (i=0;i<20;i++)
           {
               lineBuffer[i].angle = -lineBuffer[i].angle;	//mirror the angle of all the lines (to flip the track)
               
               }
           setVolumeRatio(0.25);				//set the volume back to normal
           
           if (trackShape==FLATTENING_WAITING) {
                                               trackShape = FLATTENING;   //if the tack was waiting to go flat start it off now
                                               }
           
           
}



//rotate the track a bit further upside down
void rotateZTrack()
{
           
           if (trackRotating == 1)    //rotating left
              {
                             trackRotationAngle-= 5.0;    				//increment rotation
                              setVolumeRatio(0.25+(fabs(trackRotationAngle)/360.0));   //increment volume change
                             if (trackRotationAngle < -180)
                                 {swapTrack();}				//if already rotated the full 180 degrees swap the track completely
              }
           else if (trackRotating == -1)    //rotating right
              {
                             trackRotationAngle+= 5.0;
                              setVolumeRatio(0.25+(fabs(trackRotationAngle)/360.0));
                             if (trackRotationAngle > 180)
                                 {swapTrack();}
              }
              
          
}
           




//pre calculate sin/cos from -90 to -90 clockwise                  
void setUpSinCosArray()
{
int i;

for (i=0;i<(LEVELS*2);i++)
    {
    sinTable[i] = SIN(((180.0/(LEVELS))*i)-90.0);
    cosTable[i] = COS(((180.0/(LEVELS))*i)-90.0);
    }
    
//left-hand side of track is array pos 0, right side is array pos LEVELS

}


//update the ringbuffer used to store the players last position
void updatePlayerHistory(float XPos, float YPos)
{
          previousLast = (previousLast+1)%10;		
          previousXPos[previousLast] = XPos;
          previousYPos[previousLast] = YPos;	
                          
                          
}


//calculate the vertical velocity of the player
void calcNewPlayerUp()
{

          
          waveHeight = lineBuffer[lineEP].levels[64+(int)((playerXPos)*63)];   //calculate wave height where the player is now
          
          if (waveHeight>playerYPos)				//if the wave is higher than the player pos
          {
                playerUpThrust+= UPSPEED;                    //increase up speed
                                    }          
                                        
          else if (waveHeight<playerYPos)			//if the wave is lower than the player pos
          {
               playerUpThrust-= DOWNSPEED;
               //decrease up speed
               }
                 
}

/*creates a new line on the start of ring buffer, overwrites the last line,
and moves the start pointer along*/
void addNewLine()
{
          float cornerSmoothing = 0.0;
         
         AvVol = ((AvVol*20.0) + getVolume()*5)/21.0;  //calculate a new average volume
         waveMul = (1.0/AvVol); 			//set the wave height multiplier to automatically adjust according to volume
         if (trackShape == FLAT)       			//if the track is flat randomly decide to make it round or not
         {
             if ((rand()%1000)>996) 
             {
                trackShape = ROUNDING;
                glutTimerFunc(5000, changeToRoundTimer, 1);             /*call in 5secs to set the track to flat again*/
             }  
         }
          
          
          upDateShape();				//update the shape of the track (round or flat)
          
          lineSP = (lineSP-1);                               /*move start pointer*/
          if (lineSP == -1) lineSP = 19;
          
          lineEP = (lineSP-1);                /*calculate the new end of the ring buffer*/
          if (lineEP == -1) lineEP = 19;
                 
          
          
          lineBuffer[lineSP].red = (float)rand()/RAND_MAX;   /*choose random colour*/
          lineBuffer[lineSP].blue = (float)rand()/RAND_MAX;
          lineBuffer[lineSP].green = (float)rand()/RAND_MAX;
           
          /*fill levels array with current frequency graph*/    
          getLevels(lineBuffer[lineSP].levels);             
          
          
          /*assign current line angle*/
          cornerSmoothing = 1-getVolume();		/*smoothly move from angle to angle depending on volume*/
	  
          currentAngle = (currentAngle*(8*cornerSmoothing)+currentAngleAim)/
                          ((8*cornerSmoothing)+1);  /*weighted average*/
			  
          lineBuffer[lineSP].angle = currentAngle;
          lineBuffer[lineSP].roundness = roundness;	//take on global roundness value
          
          calcNewPlayerUp();			//calculate new player up velocity
               
               
}
               

//change the track back to flattening unless the track is rotating
void changeToRoundTimer()
{
     
     if (trackShape == ROUND)
         {   
             if (trackRotating)
             trackShape = FLATTENING_WAITING;		
             else trackShape = FLATTENING;
         }
     
}


//update the shape of the track
void upDateShape()
{
     if (trackShape == ROUNDING)
        {
                    roundness+= 0.05;
                    setVolumeRatio(roundness/4.0);
                    
                    if (roundness > 1)
                    {
                                  roundness = 1;
                                  setVolumeRatio(0.25);
                                  trackShape = ROUND;
                                  drawingBottom = TRUE;
                    }
        }
        
     else if (trackShape == FLATTENING)
        {
                    roundness-= 0.05;
                    setVolumeRatio(roundness/4.0);
                    
                    if (roundness < 0)
                    {
                                  roundness = 0;
                                  setVolumeRatio(0.0);
                                  trackShape = FLAT;
                                  drawingBottom = FALSE;
                                  setVolumeRatio(0.00);
                    }
        }
}
     

/*timer - every time it is called (i.e. once a second) a new angle (corner) is calculated
and assigned to all new lines created*/                 
void updateCorner()
{
     glutTimerFunc(1000, updateCorner, 1);             /*once a second*/      
     currentAngleAim = 2-(((float)rand()/RAND_MAX)*4);    /*calculate a random angle*/
     
     
     }

/*initialise all the lines to 0 to start*/

void setupLines()
{    int i;
     
     currentAngle =0;
     for(i=0;i<20;i++)
          {
          addNewLine();           
          } 
     
}


/*draw one of the lines (line number (array subscript), start coordinates, end coordinates)*/

void drawCurrentLine(int currentLine, float xS, float yS, float zS, float xE, float yE, float zE)
{
     float angle, currentRoundness;
     int i;
     float xBit, zBit, currentX, currentZ;
     float xpos, ypos, zpos;
     int neg = 0;
     
     
     angle = lineBuffer[currentLine].angle;     /*get line angle*/
     currentRoundness = lineBuffer[currentLine].roundness; 
     if (currentRoundness > 0.5) neg = TRUE;
    
         glRotatef(angle, 0.0, 1.0, 0.0);       /*rotate line to its angle*/
     
     
         currentX = xS;                         /*get start position for the line*/
         currentZ = zS;

         xBit = (xE-xS)/LEVELS;                 /*increment for each vertex on the line*/
         zBit = (zE-zS)/LEVELS;
         
         glBegin(GL_LINE_STRIP);                /*start the line*/
         
            glColor3f(lineBuffer[currentLine].red, /*set colour*/
            lineBuffer[currentLine].green,
            lineBuffer[currentLine].blue);
            
            /*increment through the line to place each vertex*/ 
            for (i=0; i<LEVELS; i++)
            {      
               /*place a vertex in the current coordinates, with the height info got
               from the 'levels' data element*/ 
               xpos = ((1-currentRoundness)*currentX)
                      +(sinTable[i]*(TRACK_WIDTH/3+lineBuffer[currentLine].levels[i])*currentRoundness); 
               ypos = (1-currentRoundness)*lineBuffer[currentLine].levels[i]
                      +(cosTable[i]*(TRACK_WIDTH/3+lineBuffer[currentLine].levels[i])*currentRoundness); 
               
               glVertex3f(xpos, ypos, currentZ);
               currentX+=xBit;                  /*increment the current coordinates*/
               currentZ+=zBit;                  /*by the pre calculated amounts*/
            }
            
            
        if (neg)
        {    
                currentX = xS;                         /*get start position for the line*/
                 currentZ = zS;
        
                 xBit = (xE-xS)/LEVELS;                 /*increment for each vertex on the line*/
                 zBit = (zE-zS)/LEVELS;
                 
                 glBegin(GL_LINE_STRIP);                /*start the line*/
                 
                    glColor3f(lineBuffer[currentLine].red, /*set colour*/
                    lineBuffer[currentLine].green,
                    lineBuffer[currentLine].blue);
                    
                    /*increment through the line to place each vertex*/ 
                    for (i=0; i<LEVELS; i++)
                    {      
                       /*place a vertex in the current coordinates, with the height info got
                       from the 'levels' data element*/ 
                       xpos = ((1-currentRoundness)*currentX)
                              +(sinTable[i]*(TRACK_WIDTH/3+lineBuffer[currentLine].levels[i])*currentRoundness); 
                       ypos = (1-currentRoundness)*lineBuffer[currentLine].levels[i]
                              +(cosTable[i]*(TRACK_WIDTH/3+lineBuffer[currentLine].levels[i])*currentRoundness); 
                       ypos*=-1;  
                       glVertex3f(xpos, ypos, currentZ);
                       currentX+=xBit;                  /*increment the current coordinates*/
                       currentZ+=zBit;                  /*by the pre calculated amounts*/
                    }
        }

            
        glEnd();
        glTranslatef(0.0, 0.0, -LINE_DIST); /*translate ready to draw the next line*/

     }
     
/*draw the track*/
void drawLines()
{
     int currentLine;
     int i;
     

                
    
    currentLine = (lineSP-1)%20;                /*calculate the end of the ring buffer*/
    if (currentLine<0) currentLine+= 20;        /*(the line closest to the viewer)*/
    
    /*calculate player movement*/
    sinAngle = SIN(lineBuffer[currentLine].angle+(playerAngle/2.0))*playerSpeed;
          
    /*rotate a ratio of the first lines angle, based on the OfsetT amount*/
    /*i.e OfsetT = 0 then rotate full amount, OfsetT = 1 then do not rotate*/
    /*this ensures that as the line approaches the viewer its angle decreases*/
    glRotatef((lineBuffer[currentLine].angle)*(1-OfsetT), 0.0, 1.0, 0.0);
    
    /*translate the line based on ofsetT, 1 = full move, 0 = no move*/
    glTranslatef(0.0, 0.0, OfsetT*LINE_DIST);
    
    /*draw the closest line*/
    drawCurrentLine(currentLine, -TRACK_WIDTH, 0, 0, TRACK_WIDTH, 0, 0);
    
    
     for(i=2;i<20;i++)
          {
          //draw the line next closest to us then on and on
          currentLine = (lineSP-i)%20;       /*work backward from the end of the ring buffer*/
          if (currentLine<0) currentLine+= 20;
          
          /*draw the line*/
          drawCurrentLine(currentLine, -TRACK_WIDTH, 0, 0, TRACK_WIDTH, 0, 0);
          
          
                      
          }
     
     
     }
          
     
     

/*timer - called every frame to update all the states*/
void updateFrame()
{
    
     int currentLine;
     int nextLine;
     
     glutTimerFunc(gameSpeed, updateFrame, 1); /*call again*/
     
     currentLine = (lineSP-1)%20;                /*calculate the end of the ring buffer*/
     if (currentLine<0) currentLine+= 20;
     
     nextLine = (currentLine-1)%20;                /*calculate the end of the ring buffer*/
     if (nextLine<0) nextLine+= 20;
     
     roundHere = lineBuffer[nextLine].roundness*OfsetT
                 +lineBuffer[currentLine].roundness*(1-OfsetT);
     
     updatePlayer();
     
     //if (trackShape==FLAT) setVolumeRatio(0.00);
     
     if (trackRotating) rotateZTrack();

    
     OfsetT+=OFSET_INC*playerSpeed;                       /*increase offset amount (move the track)*/
     
     
     
     if (OfsetT >= 1){
                     OfsetT = 0.0;
                     addNewLine();
                     //if (drawingBottom) addNewLine(lineBufferBottom, &lineSPBottom); 
                     }                  /*if line has passed then reset the ofset 
                                        and add a new line to the ring buffer*/
                                        
     /* handle input*/
     if (pressingLeft&&onTrack){
                       playerAngle-= 0.2;
                       if (playerAngle<-10.0) playerAngle = -10.0;
                       }
     
        
     else if (pressingRight&&onTrack){
                       playerAngle+= 0.2;
                       if (playerAngle>10.0) playerAngle = 10.0;
                       }
     /* centre the player angle back to 0*/
     else playerAngle/= 1.2;
     
     
     
                                    
     glLoadIdentity();
     glutPostRedisplay();               /*redraw the screen*/
}

void changeSpeed(int speed)
{
     if (speed == FASTER)
     {
               playerSpeed+= 0.01;
               if (playerSpeed > 1.0) playerSpeed = 1.0; //increase player speed
     }

     else if (speed == SLOWER)
     {
               playerSpeed-= 0.003;
               if (playerSpeed < 0.25) playerSpeed = 0.25;	//decrease player speed
     }  
     changeMusicFreq(44100*playerSpeed);      //reduce music speed       
}

void updatePlayer()
{
      if (waveHeight>playerYPos)
          {
                onTrack = TRUE;                    //above the track
                                    }          
                                        
    else if (waveHeight<playerYPos)
          {
               onTrack = FALSE;
               //on the track
               }
    
    /*move player position accoriding to angle of track + player angle*/
     playerXPos+= sinAngle*TRACK_WIDTH;
     
      
     if (playerXPos>1) 
            if ((roundHere<0.95)||(trackShape == FLATTENING)) 
                    {playerXPos = 1;} 
            else trackRotating = -1;
            
     else if (playerXPos<-1)
            if ((roundHere<0.95)||(trackShape == FLATTENING))
                    {playerXPos = -1;}
            else trackRotating = 1;
     
    
    playerYPos += playerUpThrust;
    
    playerYPos*= 0.99; 
    
     
    
    if ((playerXPos<(-difficulty))&&(!roundHere>0.95)) changeSpeed(SLOWER);
    else if ((playerXPos>(difficulty))&&(!roundHere>0.95)) changeSpeed(SLOWER);
    else changeSpeed(FASTER);   
     
}


/*draw the player*/

void drawPlayer()
{
    float volume = getVolume()/2.0;
    int i, j = 0;
    float xpos, ypos;
    
    
    

    // if the track is round it is more complicated to work out the y position
    
    xpos = playerXPos*TRACK_WIDTH*(1-roundHere)
          +(sinTable[(int)(((playerXPos+1)/2)*LEVELS)])*(TRACK_WIDTH/3)*roundHere;  
	  

    ypos = playerYPos*(1-roundHere)
          +(cosTable[(int)(((playerXPos+1)/2)*LEVELS)])*(TRACK_WIDTH/3)*roundHere;
          
    updatePlayerHistory(xpos, ypos);
    
    glPushMatrix();
    
    glColor3f(1.0-onTrack, 0.8, 1.0);
    glTranslatef(xpos,ypos,-LINE_DIST);
    glutSolidSphere(0.01+volume,		//draw a sphere for the player, size based on volume
                     16, 16);
    glPopMatrix();
    
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    for (i=0;i<10;i++)			//draw a the trail left by the player
    {
            j = (previousLast - i);
            if (j<0) j+= 10;
            glPushMatrix();
            glColor4f(1.0-onTrack, 0.8, 1.0, (0.4-(i/10.0)));
            glTranslatef(previousXPos[j],previousYPos[j],(-LINE_DIST)+i*0.10*playerSpeed);
            glutSolidSphere(((0.01+volume) - (i/10)*0.01),
                     16, 16);
            glPopMatrix();
            }
    glDisable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
}


//not actually a menu, code for displaying the small 'altitude' graph top right hand corner

void drawMenu()
{
          
 //draw an altitude graph         
glBegin(GL_LINE_STRIP);
      glColor3f(1.0, 1.0, 1.0);
      
      glVertex3f(1.3, 1.5, -5);
      glVertex3f(1.3, 1.0, -5);
      glVertex3f(1.4, 1.0, -5);
      glVertex3f(1.4, 1.5, -5);
      glVertex3f(1.3, 1.5, -5);
glEnd();

glBegin(GL_LINE_STRIP);
      glColor3f(1.0, 1.0, 1.0);
      
      glVertex3f(1.3, 1.25+waveHeight, -5);
      glVertex3f(1.4, 1.25+waveHeight, -5);

glEnd();


glBegin(GL_LINE_STRIP);
      glColor3f(1.0, 0.0, 1.0);
      
      glVertex3f(1.3, 1.25+(playerYPos), -5);
      glVertex3f(1.4, 1.25+(playerYPos), -5);

glEnd();
        
          
}



/*draw function*/

void display(void)
{
    
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glLoadIdentity();
    
    drawMenu();
    
    float playerPosOnScreen;
    
    if (roundHere>0.9) playerPosOnScreen = 0.0;
    else playerPosOnScreen = (playerXPos*TRACK_WIDTH)/2.0;
         
     /*position the camera*/
     
     // update the camera position smoothly by taking an average of old position and new position
     newCam.eyeX = ((oldCam.eyeX*10)+playerPosOnScreen)/11.0;
     newCam.eyeY = (((oldCam.eyeY)*10)+(playerYPos))/11.0;
     
     newCam.lookX = newCam.eyeX;
     newCam.lookY = newCam.eyeY;
     
     oldCam = newCam;
     
     
     
               
     gluLookAt( newCam.eyeX,     /*eye*/
                0.8 + newCam.eyeY,
                2.0,
                
                newCam.lookX,     /*look at*/
                0.3 + newCam.lookY,
                0.0,
                
                0.0,     /*up*/
                1.0,
                0.0);
                
    glRotatef(trackRotationAngle, 0, 0, 1.0);
                
    glRotatef(playerAngle, 0.0, 1.0, 0.0);
    
    glPushMatrix();
    /*draw the track*/
    drawLines();
    glPopMatrix();
    
    /*draw the player*/
    drawPlayer();
    


    glutSwapBuffers();
}



//set flags based on input
void keyUp(int key, int x, int y)
{
    if (key == GLUT_KEY_LEFT) pressingLeft = 0;
    else if (key == GLUT_KEY_RIGHT) pressingRight = 0;
}



void key(int key, int x, int y)
{
    if (key == GLUT_KEY_LEFT) pressingLeft = 1;
    else if (key == GLUT_KEY_RIGHT) pressingRight = 1;
}



static void idle(void)
{
    glutPostRedisplay();
}



/* Program entry point */


int main(int argc, char *argv[])
{
    if (argc > 1) {
             difficulty = atof(argv[1]);
             gameSpeed = 20.0-(atof(argv[1])*10.0);};  //accept a 'difficulty' argument and adjust gamespeed appropriately
    
    difficulty = 0.7 - (difficulty/2.0);
    
    glutInit(&argc, argv);
    srand( (unsigned)time( NULL ) );	//generate a random seed
    setUpSinCosArray();
    glutInitWindowSize(640,480);	//set up the glut window
    glutInitWindowPosition(10,10);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);

    glutCreateWindow("SoundTrack");
    
    glViewport(0, 0, 640, 480);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(35.0, 1.0, 0.1, 50.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity() ;

    glutDisplayFunc(display);
    glutSpecialFunc(key);
    glutSpecialUpFunc(keyUp);
    glutIdleFunc(idle);

    glClearColor(0,0,0,0);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    glEnable(GL_DEPTH_TEST);
    
    initMusic();		//initialise the music

    setupLines();		//initialise the track
    
    
    
    glutTimerFunc(2000, updateCorner, 1);  //start game loop timers
    
    glutTimerFunc(10, updateFrame, 1);
    
    
    
    
    glutMainLoop();		//enter main loop and wait for interupts
    

    return EXIT_SUCCESS;
}

